@@ -3,6 +3,7 @@ import importlib.util
 import json
 import sys
 import types
+import warnings
 from collections.abc import MutableMapping as GenericMapping
 
 import toml
@@ -15,6 +16,8 @@ from typing import Mapping, MutableMapping, Callable
 
 from beacon import xyz
 
+ALLOWED_EXTS = ('.yaml', '.yml', '.json', '.toml', '.xyz')
+
 
 # decorate internal methods in ADict
 def mutate_attribute(fn):
@@ -158,7 +161,7 @@ class ADict(Dict):
         if isinstance(names, str):
             if names in self._data:
                 value = self._data[names]
-            elif self._is_default_defined:
+            elif self._is_default_defined and not self.frozen:
                 value = self.get_default()
                 self._data[names] = value
             else:
@@ -220,7 +223,7 @@ class ADict(Dict):
     def __getstate__(self):
         state = dcp(self.__dict__)
         state.pop('_mutate_attribute')
-        return self.__dict__
+        return state
 
     @mutate_attribute
     def __setstate__(self, state):
@@ -249,6 +252,8 @@ class ADict(Dict):
         if name in self:
             return self.__getitem__(name)
         elif self._is_default_defined:
+            if self._frozen:
+                raise KeyError(f'The key "{name}" does not exist.')
             return self.get_default()
         else:
             return default
@@ -469,7 +474,7 @@ class ADict(Dict):
             ext = os.path.splitext(path)[1].lower()
             if ext in ('.yml', '.yaml'):
                 with open(path, 'rb') as f:
-                    self._data = yaml.load(f, Loader=yaml.FullLoader, **kwargs)
+                    self._data = yaml.load(f, Loader=yaml.FullLoader)
             elif ext == '.json':
                 with open(path, 'r') as f:
                     self._data = json.load(f, **kwargs)
@@ -486,6 +491,63 @@ class ADict(Dict):
     def load_mm_config(self, path):
         self.update(self.from_mm_config(path).to_dict())
 
+    @classmethod
+    def compose_hierarchy(
+        cls,
+        root,
+        config_filename='config',
+        select=None,
+        overrides=None,
+        on_missing='error',
+        required=None
+    ):
+        select = select or {}
+        loaded_paths = []
+        config = cls()
+
+        def _load_first(_base_path, *, label):
+            _is_missing = True
+            for _ext in ALLOWED_EXTS:
+                _config_path = os.path.join(root, _base_path+_ext)
+                if os.path.exists(_config_path):
+                    config.update(cls.from_file(_config_path))
+                    loaded_paths.append((label, _config_path))
+                    _is_missing = False
+            return _is_missing
+
+        if not _load_first(config_filename, label='base') and on_missing == 'error':
+            raise FileNotFoundError(f'{config_filename}.[yaml|yml|json|toml|xyz] not found in {root}')
+        for group, options in select.items():
+            options = options if isinstance(options, (list, tuple)) else [options]
+            for option in options:
+                is_missing = _load_first(os.path.join(group, str(option)), label=f'{group}:{option}')
+                if is_missing:
+                    if on_missing == 'error':
+                        raise FileNotFoundError(f'{group}/{option} not found')
+                    elif on_missing == 'warn':
+                        warnings.warn(f'missing: {group}/{option}')
+        for key, value in (overrides or {}).items():
+            if isinstance(value, Mapping):
+                value = ADict(**value)
+            current_config = config
+            sub_keys = key.split('.')
+            for sub_key in sub_keys[:-1]:
+                if sub_key not in current_config or not isinstance(current_config[sub_key], ADict):
+                    current_config[sub_key] = ADict()
+                current_config = current_config[sub_key]
+            current_config[sub_keys[-1]] = value
+        for is_required in (required or []):
+            current_config = config
+            is_missing = False
+            for key in is_required.split('.'):
+                if key not in current_config:
+                    is_missing = True
+                    break
+                current_config = current_config[key]
+            if is_missing:
+                raise KeyError(f"Missing required key: {is_required}")
+        return config
+
     def dump(self, path, **kwargs):
         dir_path = os.path.dirname(os.path.realpath(path))
         os.makedirs(dir_path, exist_ok=True)
@@ -493,7 +555,7 @@ class ADict(Dict):
         if ext in ('.yml', '.yaml'):
             with open(path, 'w') as f:
                 return yaml.dump(self.to_dict(), f, Dumper=yaml.Dumper, **kwargs)
-        elif ext == 'toml':
+        elif ext == '.toml':
             with open(path, 'w') as f:
                 return toml.dump(self.to_dict(), f, **kwargs)
         elif ext == '.json':
@@ -512,4 +574,9 @@ class ADict(Dict):
                 raise KeyError(f'The key {src_key} does not exist.')
         self.__setitem__(tgt_keys, [self._data.pop(src_key) for src_key in src_keys])
 
+    @classmethod
+    def auto(cls):
+        return cls(default=lambda: cls.auto())
+
+
 
