@@ -1,5 +1,7 @@
 import argparse
+import hashlib
 import inspect
+import pickle
 import sys
 import uuid
 import warnings
@@ -171,6 +173,26 @@ def _print_config(config):
     sys.exit(0)
 
 
+def _get_func_trace_id(func):
+    return f'{func.__module__}.{func.__qualname__}'
+
+
+def _generate_func_fingerprint(func, trace_id=None):
+    code_obj = func.__code__
+    code_info = (
+        code_obj.co_code,
+        code_obj.co_consts,
+        code_obj.co_names,
+        code_obj.co_varnames,
+        code_obj.co_filename,
+        code_obj.co_name,
+        code_obj.co_firstlineno
+    )
+    trace_id = _get_func_trace_id(func) if trace_id is None else trace_id
+    code_hash = hashlib.sha256(repr(code_info).encode()).hexdigest()
+    return ADict(**{trace_id: code_hash})
+
+
 class Scope:
     registry = ADict()
     parsed = False
@@ -201,6 +223,7 @@ class Scope:
         self.config_in_compute = None
         self.mode = 'ON'
         self.is_applied = False
+        self._traced_data = ADict(fingerprints=ADict())
 
     def activate(self):
         self.mode = 'ON'
@@ -213,7 +236,28 @@ class Scope:
         self.deactivate()
         yield
         self.activate()
-    
+
+    def trace(self, trace_id=None):
+        def decorator(func):
+            self._traced_data.fingerprints.update(_generate_func_fingerprint(func, trace_id=trace_id))
+            return self(func)
+        return decorator
+
+    def runtime_trace(self, init_fn=None, inspect_fn=None, trace_id=None):
+        def decorator(func):
+            def inner(*args, **kwargs):
+                nonlocal trace_id
+                if init_fn is not None:
+                    init_fn()
+                results = inspect_results = self(func)(*args, **kwargs)
+                if inspect_fn is not None:
+                    inspect_results = inspect_fn(results)
+                trace_id = _get_func_trace_id(func) if trace_id is not None else trace_id
+                inspect_hash = hashlib.sha256(pickle.dumps(inspect_results)).hexdigest()
+                self._traced_data.fingerprints.update({trace_id: inspect_hash})
+            return inner
+        return decorator
+
     def register(self):
         registry = self.__class__.registry
         if len(registry) == 0:
@@ -427,10 +471,6 @@ class MultiScope:
     def __init__(self, *scopes):
         self.scopes = scopes
         self.register_all()
-        if Scope.parsed:
-            for scope in self.scopes:
-                scope.reset_user_inputs()
-            parse_args_pythonic()
 
     def register_all(self):
         Scope.registry.clear()
@@ -439,6 +479,8 @@ class MultiScope:
 
     def __call__(self, func):
         def decorator(*args, **kwargs):
+            # if not Scope.parsed:
+            parse_args_pythonic()
             arguments = inspect.getfullargspec(func)
             name_spaces = set(arguments.args+arguments.kwonlyargs)
             for scope in self.scopes:
